import numpy as np


class ReferenceImputer:
    '''
    Impute features using reference values.

    Args:
      reference: the reference value for replacing missing features.
    '''
    def __init__(self, reference):
        self.reference = reference
        self.samples = 1

    def __call__(self, x, S):
        x_ = x.copy()
        x_[~S] = self.reference.repeat(len(x), 1)[~S]


class MarginalImputer:
    '''
    Impute features using a draw from the joint marginal.

    Args:
      data: np.ndarray of size (samples, dimensions) representing the data
        distribution.
      samples: number of samples to draw from marginal distribution.
    '''
    def __init__(self, data, samples):
        self.data = data
        self.samples = samples
        self.N = len(data)
        self.x_addr = None
        self.x_repeat = None

    def __call__(self, x, S):
        if self.x_addr == id(x):
            x = self.x_repeat
        else:
            self.x_addr = id(x)
            x = np.repeat(x, self.samples, 0)
            self.x_repeat = x
        S = np.repeat(S, self.samples, 0)
        samples = self.data[np.random.choice(self.N, len(x), replace=True)]
        x_ = x.copy()
        x_[~S] = samples[~S]
        return x_



import pickle
import numpy as np
import matplotlib.pyplot as plt


class SAGE:
    '''For storing and plotting SAGE values.'''
    def __init__(self, values, std):
        self.values = values
        self.std = std

    def plot(self,
             feature_names,
             sort_features=True,
             max_features=np.inf,
             figsize=(10, 7),
             orientation='horizontal',
             error_bars=True,
             color='tab:green',
             title='Feature Importance',
             title_size=20,
             tick_size=16,
             label_size=16,
             label_rotation=None,
             return_fig=False):
        '''
        Plot SAGE values.

        Args:
          feature_names: list of feature names.
          sort_features: whether to sort features by their SAGE values.
          max_features: number of features to display.
          fig: matplotlib figure (optional).
          figsize: figure size (if fig is None).
          orientation: horizontal (default) or vertical.
          error_bars: whether to include standard deviation error bars.
          color: bar chart color.
          title: plot title.
          title_size: font size for title.
          tick_size: font size for feature names and numerical values.
          label_size: font size for label.
          label_rotation: label rotation (for vertical plots only).
          return_fig: whether to return matplotlib figure object.
        '''
        fig = plt.figure(figsize=figsize)
        ax = fig.gca()

        # Sort features if necessary.
        if len(feature_names) > max_features:
            sort_features = True
        values = self.values
        std = self.std
        if sort_features:
            argsort = np.argsort(values)[::-1]
            values = values[argsort]
            std = std[argsort]
            feature_names = np.array(feature_names)[argsort]

        # Remove extra features if necessary.
        if len(feature_names) > max_features:
            feature_names = list(feature_names[:max_features]) + ['Remaining']
            values = (list(values[:max_features])
                      + [np.sum(values[max_features:])])
            std = (list(std[:max_features])
                   + [np.sum(std[max_features:] ** 2) ** 0.5])

        # Make plot.
        if orientation == 'horizontal':
            # Bar chart.
            if error_bars:
                ax.barh(np.arange(len(feature_names))[::-1], values,
                        color=color, xerr=std)
            else:
                ax.barh(np.arange(len(feature_names))[::-1], values,
                        color=color)

            # Feature labels.
            if label_rotation is not None:
                raise ValueError('rotation not supported for horizontal charts')
            ax.set_yticks(np.arange(len(feature_names))[::-1])
            ax.set_yticklabels(feature_names, fontsize=label_size)

            # Axis labels and ticks.
            ax.set_ylabel('')
            ax.set_xlabel('SAGE value', fontsize=label_size)
            ax.tick_params(axis='x', labelsize=tick_size)

        elif orientation == 'vertical':
            # Bar chart.
            if error_bars:
                ax.bar(np.arange(len(feature_names)), values, color=color,
                       yerr=std)
            else:
                ax.bar(np.arange(len(feature_names)), values, color=color)

            # Feature labels.
            if label_rotation is None:
                label_rotation = 90
            if label_rotation < 90:
                ha = 'right'
                rotation_mode = 'anchor'
            else:
                ha = 'center'
                rotation_mode = 'default'
            ax.set_xticks(np.arange(len(feature_names)))
            ax.set_xticklabels(feature_names, rotation=label_rotation, ha=ha,
                               rotation_mode=rotation_mode,
                               fontsize=label_size)

            # Axis labels and ticks.
            ax.set_ylabel('SAGE value', fontsize=label_size)
            ax.set_xlabel('')
            ax.tick_params(axis='y', labelsize=tick_size)

        else:
            raise ValueError('orientation must be horizontal or vertical')

        # Remove spines.
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)

        ax.set_title(title, fontsize=title_size)
        plt.tight_layout()

        if return_fig:
            return fig
        else:
            return

    def save(self, filename):
        '''Save SAGE object.'''
        if isinstance(filename, str):
            with open(filename, 'wb') as f:
                pickle.dump(self, f)
        else:
            raise TypeError('filename must be str')


def load(filename):
    '''Load SAGE object.'''
    with open(filename, 'rb') as f:
        sage_values = pickle.load(f)
        if isinstance(sage_values, SAGE):
            return sage_values
        else:
            raise ValueError('object is not instance of SAGE class')



from sage import utils, core, imputers
from .core import load, SAGE
from .imputers import MarginalImputer, ReferenceImputer
from .permutation_sampler import PermutationSampler
from .iterated_sampler import IteratedSampler



import setuptools

setuptools.setup(
    name="sage-importance",
    version="0.0.2",
    author="Ian Covert",
    author_email="icovert@cs.washington.edu",
    description="For calculating global feature importance using Shapley values.",
    long_description="""
        SAGE (Shapley Additive Global importancE) is a game theoretic approach 
        for understanding black-box machine learning models. It summarizes the 
        importance of each feature based on the predictive power it 
        contributes and accounts for complex interactions by using the Shapley 
        value from cooperative game theory. See the 
        [GitHub page](https://github.com/iancovert/sage/) for examples, and see 
        the [paper](https://arxiv.org/abs/2004.00668) for more details.
    """,
    long_description_content_type="text/markdown",
    url="https://github.com/iancovert/sage/",
    packages=['sage'],
    install_requires=[
        'numpy',
        'matplotlib',
        'tqdm'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering"
    ],
    python_requires='>=3.6',
)
